#  Copyright (c) PARRRATE T&V 2021. All rights reserved.

from bu4.evaluation.av.envtype import envtype
from bu4.evaluation.constructs.edelayed import EDelayed
from bu4.evaluation.constructs.elambda import ELambda
from bu4.evaluation.constructs.evaluable import Evaluable
from bu4.indexing.constructs.idelayed import IDelayed
from bu4.indexing.constructs.ilambda import ILambda
from bu4.indexing.states.afterindexing import AfterIndexing
from bu4.indexing.states.indexingfinished import IndexingFinished
from bu4.indexing.states.indexinglinked import IndexingLinked
from bu4.indexing.states.indexingstate import IndexingState
from bu4.indexing.targets.ailambda import AILambda
from bu4.linking.constructs.linked import Linked

__all__ = ('LLambda',)


class LLambda(Linked):
    value: Linked

    def __init__(self, name: bytes, value: Linked):
        self.name = name
        self.value = value
        self.future = self.value.future - {name}
        self.multifuture = self.future
        self.used = name in value.future
        self.memoize = name in value.multifuture

    def evaluable(self, env: envtype) -> Evaluable:
        return (
            ELambda(
                {name: env[name] for name in self.future} if self.used else env,
                self.name,
                self.value
            )
            if
            self.used
            else
            EDelayed(env, self.value)
        )

    def index(self, promise: list[bytes]) -> IndexingState:
        if not self.used:
            return AfterIndexing(
                IndexingLinked(promise, self.value),
                AILambda(
                    lambda value: IndexingFinished(IDelayed(value))
                )
            )
        else:
            table = [i for i in range(len(promise)) if promise[i] in self.future]
            return AfterIndexing(
                IndexingLinked([self.name] + [promise[i] for i in table], self.value),
                AILambda(
                    lambda value: IndexingFinished(ILambda(
                        value,
                        table,
                        memoize=self.memoize
                    ))
                )
            )

    def __str__(self):
        return f'({self.name.decode()}){self.value}'
